//RESTROE3 -   cont version of Restroe2, GV2002-2007
//Orbital elements <-> Orbital Elements of TwoBody Configuration
//Restroe3 was created at 12/2/2007 


#include <math.h>
#include <nbody\restroe3.h>


#define RTINY 1.0e-15
//internal globals
double tmu, mBody, mSun, A_metric = 1;
int RPLANAR = 0, MODEPIPI = 0, BARYCENTRICDATA = 1;



void InitRestroe(int planar, double mbody, double msun, int barycentricdata, double ametric, int modepipi)
//mbody,msun: masses of the two bodies (body and sun)
//ametric distance SUN-Secondary body (for normalized units use ametric=1
//planar: set TRUE(1) when the body data vector x is (t,x,y,xdot,ydot) 
//			or FALSE(0) for 3D case where x=(t,x,y,z,xdot,ydot,zdot)
//modepipi=true : angles output interval is (-pi,pi) or (-180,180)  
{
	mBody = mbody; mSun = msun;
	tmu = mbody + msun;
	A_metric = ametric;
	RPLANAR = planar;
	MODEPIPI = modepipi;
	BARYCENTRICDATA = barycentricdata;
}

//simple calls for normal-use configurations
//1. Barycentric
void InitRestroeB(int planar, double mbody, double msun)
{
	mBody = mbody; mSun = msun; tmu = mbody + msun; RPLANAR = planar;
	BARYCENTRICDATA = 1; 	MODEPIPI = 0; A_metric = 1.0;
}
//2. Heliocentric
void InitRestroeH(int planar, double mbody, double msun)
{
	mBody = mbody; mSun = msun; tmu = mbody + msun; RPLANAR = planar;
	BARYCENTRICDATA = 0; 	MODEPIPI = 0; A_metric = 1.0;
}
//restricted_default call
void InitRestroeDef(int planar, double ametric)
{
	mBody = 0; mSun = 1; tmu = 1;
	A_metric = ametric; RPLANAR = planar;
	MODEPIPI = 0; BARYCENTRICDATA = 0;
}

//==============================================================================

void GetOrbitalElements(double x[], OEPOS *q, int DISPLAYVALUES)
//Murray&Dermott formulae
//x[]:input data in Heliocentric coordinate system
//q: Structure of orbital elements to be returned
//DISPLAYVALUES : format of OE (1: display mode (deg, AU), 0: calc mode (rad, nodim axis)
{

	double xa[7];
	SwitchToSystemDimension(1, x, xa);
	if (BARYCENTRICDATA) BarycentricToHeliocentric(xa);
	double r2 = xa[1] * xa[1] + xa[2] * xa[2] + xa[3] * xa[3];
	double r = sqrt(r2);
	double V2 = xa[4] * xa[4] + xa[5] * xa[5] + xa[6] * xa[6];
	double hx = xa[2] * xa[6] - xa[3] * xa[5];
	double hy = xa[3] * xa[4] - xa[1] * xa[6];
	double hz = xa[1] * xa[5] - xa[2] * xa[4];
	double h2 = hx*hx + hy*hy + hz*hz;
	double h = sqrt(hx*hx + hy*hy + hz*hz);
	double ia = 2 / r - V2 / tmu;
	q->a = 1 / ia;
	q->e = 1 - h2*ia / tmu; if (q->e>0) q->e = sqrt(q->e); else q->e = 1; //hyper orbit
	q->meanmotion = sqrt(tmu*ia*ia*ia);
	q->i = acos(hz / h); if (fabs(q->i)<RTINY) q->i = 0;

	//OMEGA
	if (q->i == 0.0 || fabs(hx)<RTINY) q->Omega = 0.0;
	else q->Omega = atan2(hx, -hy) + pi2;
	//f+omega
	double sinomf, cosomf;
	if (q->i == 0)
	{
		sinomf = (xa[2] * cos(q->Omega) - xa[1] * sin(q->Omega)) / r;
		cosomf = (xa[1] * cos(q->Omega) + xa[2] * sin(q->Omega)) / r;
	}
	else
	{
		sinomf = xa[3] / (r*sin(q->i));
		cosomf = (xa[1] * cos(q->Omega) + xa[2] * sin(q->Omega)) / r;
	}
	double omf = atan2(sinomf, cosomf) + pi2; //omega+f
											  //f : from Murray, page 53
	double ee = q->e;
	if (ee<RTINY) q->f = 0;
	else
	{
		double p = q->a*(1 - ee*ee);
		double rrdot = xa[1] * xa[4] + xa[2] * xa[5] + xa[3] * xa[6];
		double Rdot = V2 - h2 / r2; if (Rdot<RTINY) Rdot = 0;
		Rdot = sqrt(Rdot);
		if (rrdot<0) Rdot = -Rdot;
		double fsin = p*Rdot / (h*ee);
		double fcos = (p / r - 1) / ee;
		q->f = atan2(fsin, fcos) + pi2; q->f = fmod(q->f, pi2);
	}
	//omega,omegabar	
	q->omega = omf - q->f + pi2;
	q->omegabar = q->omega + q->Omega + pi2;
	q->omega = fmod(q->omega, pi2);
	q->Omega = fmod(q->Omega, pi2);
	q->omegabar = fmod(q->omegabar, pi2);
	//M,lambda,E
	q->M = GetMeanFromTrue(q->f, q->e);  q->M = fmod(q->M + pi2, pi2);
	q->lamda = fmod(q->M + q->omegabar, pi2);
	q->E = GetEccentricFromTrue(q->f, q->e); q->E = fmod(q->E + pi2, pi2);
	//tau
	q->tau = q->M / q->meanmotion;
	//Units 
	if (MODEPIPI) {
		MPI_PI(&q->E); MPI_PI(&q->i); MPI_PI(&q->f); MPI_PI(&q->M); MPI_PI(&q->tau);
		MPI_PI(&q->lamda); MPI_PI(&q->omega); MPI_PI(&q->Omega); MPI_PI(&q->omegabar);
	}
	if (DISPLAYVALUES) OE_DisplayValues(q);
}

//=================================================

void GetInertialFromOrbital(OEDEF q, double x[])
//Roy, page 102
//returns the inertial position of body. 
//!!!Attention: if OE refers to heliocentric system, output refers to qoordinates where SUN is at zero  
{
	double xx[7];
	if (q.modedisplay) OE_CalcValues(&q);
	//Calc Eccentric anomaly from inputangle
	double E, M;
	switch (q.inputangle) {
	case 0: E = q.elpos; break;
	case 1: E = GetEccentricFromM(q.elpos, q.e); break;
	case 2: E = GetEccentricFromTrue(q.elpos, q.e); break;
	case 3: M = q.elpos - (q.omega + q.Omega); E = GetEccentricFromM(M, q.e); break;
	case 4: M = MeanMotion(q.a)*q.elpos; E = GetEccentricFromM(M, q.e); break;
	default: E = 0;
	}
	//----------------------------------------
	if (q.i<0) q.i *= -1.0; //inclination belongs to [0,pi]
							//see Roy page 102.
	double co = cos(q.omega); double so = sin(q.omega);
	double cO = cos(q.Omega); double sO = sin(q.Omega);
	double ci = cos(q.i); double si = sin(q.i);
	double l1 = cO*co - sO*so*ci;
	double m1 = sO*co + cO*so*ci;
	double n1 = so*si;
	double l2 = -cO*so - sO*co*ci;
	double m2 = -sO*so + cO*co*ci;
	double n2 = co*si;
	double b = q.a*sqrt(1 - q.e*q.e);
	//------------------------------------------
	double cE = cos(E);
	double sE = sin(E);
	double r = q.a*(1 - q.e*cE);
	double n = sqrt(tmu / (q.a*q.a*q.a));
	double nar = n*q.a / r;
	//-----------------------------------------
	xx[0] = 0;
	xx[1] = q.a*l1*cE + b*l2*sE - q.a*q.e*l1;
	xx[2] = q.a*m1*cE + b*m2*sE - q.a*q.e*m1;
	xx[3] = q.a*n1*cE + b*n2*sE - q.a*q.e*n1;
	xx[4] = nar*(b*l2*cE - q.a*l1*sE);
	xx[5] = nar*(b*m2*cE - q.a*m1*sE);
	xx[6] = nar*(b*n2*cE - q.a*n1*sE);
	SwitchToSystemDimension(-1, xx, x);
	if (BARYCENTRICDATA) HeliocentricToBarycentric(x);
}

//====== Transformations ================

void OEPOStoOEDEF(OEPOS qo, OEDEF *qd)
{
	qd->a = qo.a; qd->e = qo.e; qd->i = qo.i;
	qd->omega = qo.omega; qd->Omega = qo.Omega;
	qd->inputangle = 1; qd->elpos = qo.M;
	qd->modedisplay = qo.modedisplay;
}

OEDEF oedef(OEPOS qo) //direct
{
	OEDEF qd;
	qd.a = qo.a; qd.e = qo.e; qd.i = qo.i;
	qd.omega = qo.omega; qd.Omega = qo.Omega;
	qd.inputangle = 1; qd.elpos = qo.M;
	qd.modedisplay = qo.modedisplay;
	return qd;
}

void OEDEFtoOEPOS(OEDEF qd, OEPOS *qo)
{
	int  mdisplay = qd.modedisplay;
	if (mdisplay) OE_CalcValues(&qd);
	qo->a = qd.a; qo->e = qd.e; qo->i = qd.i; qo->meanmotion = MeanMotion(qd.a);
	qo->omega = qd.omega; qo->Omega = qd.Omega;
	qo->omegabar = qd.omega + qd.Omega; qo->omegabar = fmod(2 * pi2 + qo->omegabar, pi2);
	switch (qd.inputangle) {
		//0:E, 1=M, 2=f, 3=lambda 4:tau
	case 0: qo->E = qd.elpos;
		qo->M = qo->E - qo->e*qo->E; qo->tau = qo->M / qo->meanmotion;
		qo->lamda = qo->M + qo->omegabar; 	qo->lamda = fmod(2 * pi2 + qo->lamda, pi2);
		qo->f = GetTrueFromEccentric(qo->E, qo->e);
		break;
	case 1: qo->M = qd.elpos;
		qo->E = GetEccentricFromM(qo->M, qo->e);
		qo->tau = qo->M / qo->meanmotion;
		qo->lamda = qo->M + qo->omegabar; 	qo->lamda = fmod(2 * pi2 + qo->lamda, pi2);
		qo->f = GetTrueFromEccentric(qo->E, qo->e);
		break;
	case 2: qo->f = qd.elpos;
		qo->E = GetEccentricFromTrue(qo->f, qo->e);
		qo->M = qo->E - qo->e*qo->E;
		qo->tau = qo->M / qo->meanmotion;
		qo->lamda = qo->M + qo->omegabar; 	qo->lamda = fmod(2 * pi2 + qo->lamda, pi2);
		break;
	case 3: qo->lamda = qd.elpos;
		qo->M = qo->lamda - qo->omegabar; qo->M = fmod(2 * pi2 + qo->M, pi2);
		qo->E = GetEccentricFromTrue(qo->f, qo->e);
		qo->tau = qo->M / qo->meanmotion;
		qo->f = GetTrueFromEccentric(qo->E, qo->e);
		break;
	case 4: qo->tau = qd.elpos;
		qo->M = qo->tau*qo->meanmotion;
		qo->E = GetEccentricFromM(qo->M, qo->e);
		qo->lamda = qo->M + qo->omegabar; 	qo->lamda = fmod(2 * pi2 + qo->lamda, pi2);
		qo->f = GetTrueFromEccentric(qo->E, qo->e);
	}
	if (MODEPIPI) {
		MPI_PI(&qo->E); MPI_PI(&qo->f); MPI_PI(&qo->M); MPI_PI(&qo->tau);
		MPI_PI(&qo->lamda); MPI_PI(&qo->omega); MPI_PI(&qo->Omega); MPI_PI(&qo->omegabar);
	}
	if (mdisplay) OE_DisplayValues(qo);
}


void BarycentricToHeliocentric(double x[])
{
	double ratio = tmu / mSun;
	for (int i = 0; i<7; i++) x[i] *= ratio;
}
void HeliocentricToBarycentric(double x[])
{
	double ratio = mSun / tmu;
	for (int i = 0; i<7; i++) x[i] *= ratio;
}



//----------------------------------------------------
//----------auxiliary functions  1 -------------------
//----------------------------------------------------

double MeanMotion(double a)
{
	return sqrt(tmu / (a*a*a));
}

double GetEccentricFromTrue(double f, double e)  //rad
{
	f += pi2; f = fmod(f, pi2);
	if (fabs(f - pi)<RTINY) return pi;
	double s = sqrt((1.0 - e) / (1.0 + e));
	double E = 2.0*atan(tan(f / 2)*s);
	if (E<0) E += pi2;
	return E;
}

double GetTrueFromEccentric(double E, double e)  //rad
{
	E = fmod(E + pi2, pi2);
	if (fabs(E - pi)<RTINY) return pi;
	double s = sqrt((1.0 + e) / (1.0 - e));
	double f = 2.0*atan(tan(E / 2)*s);
	if (f<0) f += pi2;
	return f;
}


double GetMeanFromTrue(double f, double e)  //rad
											//use Eccentric anomaly as intermediate
{
	double E = GetEccentricFromTrue(f, e);
	return E - e*sin(E); //=M
}


double GetEccentricFromM(double M, double e) //M in rad, Newton-Raphson
{
	M = fmod(M, pi2);
	double E0;
	if (fabs(M)<RTINY) return 0;
	if (fabs(M - pi)<RTINY) return pi;
	if (fabs(M + pi)<RTINY) return -pi;
	E0 = M + 0.85*e*sin(M) / fabs(sin(M));
	double E = E0; int k = 0;
	do
	{
		E0 = E; k++;
		E = E0 - (E0 - e*sin(E0) - M) / (1 - e*cos(E0));
	} while ((fabs(E0 - E)>RTINY) && k<32);
	return E;
}




//----------------------------------------------------
//----------auxiliary functions  2 -------------------
//----------------------------------------------------


void MPI_PI(double *x)
{
	(*x) = fmod((*x), pi2);
	if (*x>pi) (*x) -= pi2;
	if (*x<-pi) (*x) += pi2;
}

void M180_180(double *x)
{
	(*x) = fmod((*x), 360);
	if ((*x)>180) *x -= 360;
	if ((*x)<-180) *x += 360;
}


void OE_DisplayValues(OEPOS *q)  //deg units and a in au
{
	double ff = 180 / pi; q->a *= A_metric;
	(q->i) *= ff;	(q->f) *= ff; (q->Omega) *= ff; (q->omega) *= ff;
	(q->M) *= ff; (q->lamda) *= ff; (q->omegabar) *= ff; (q->E) *= ff; (q->tau) *= ff;
	q->modedisplay = 1;
}
void OE_CalcValues(OEPOS *q) //rad units and a in nodim units
{
	double ff = pi / 180; q->a /= A_metric;
	(q->i) *= ff;	(q->f) *= ff; (q->Omega) *= ff; (q->omega) *= ff;
	(q->M) *= ff; (q->lamda) *= ff; (q->omegabar) *= ff; (q->E) *= ff; (q->tau) *= ff;
	q->modedisplay = 0;
}
void OE_DisplayValues(OEDEF *q)  //deg units and a in au
{
	double ff = 180 / pi; q->a *= A_metric;
	(q->i) *= ff;	(q->elpos) *= ff; (q->Omega) *= ff; (q->omega) *= ff;
	q->modedisplay = 1;
}
void OE_CalcValues(OEDEF *q) //rad units and a in nodim units
{
	double ff = pi / 180; q->a /= A_metric;
	(q->i) *= ff;	(q->elpos) *= ff; (q->Omega) *= ff; (q->omega) *= ff;
	q->modedisplay = 0;
}


//------------------------------------------------------
//-------------- auxiliary 3 ---------------------------

void SwitchToSystemDimension(int direction, double x[], double xa[])
//according to PLANAR or not mode
//direction= 1 the x[5] pos-vel vector is copied in xa[7] pos-vel vector
//direction=-1 the x[7] pos-vel vector is copied in xa[5] pos-vel vector
{
	if (!RPLANAR || direction == 0) { for (int j = 0; j<7; j++) xa[j] = x[j]; return; }
	if (direction>0)
	{
		xa[0] = x[0]; xa[1] = x[1]; xa[2] = x[2]; xa[3] = 0.0; xa[4] = x[3]; xa[5] = x[4]; xa[6] = 0.0;
	}
	else
	{
		xa[0] = x[0]; xa[1] = x[1]; xa[2] = x[2]; xa[3] = x[4]; xa[4] = x[5];
	}
}


//=====================================================================
// TOOL FUNCTIONS
void fixw(double *x, int mode_deg, int mode_180_pi)
//x: I/O angle value
//mode_deg: if true units are degrees else rads
//mode_180_pi : if true values are given in [-pi,pi] or [-180,180] else [0,2pi] or [0,360]
{
	double w, W = *x;
	if (mode_deg) w = 360.0; else w = pi2;
	W += (10 * w); W = fmod(W, w);
	if (fabs(w - W)<RTINY)   W = 0;
	if (fabs(W - w / 2)<RTINY) W = w / 2;
	if (mode_180_pi&&W>w / 2) W -= w;
	*x = W;
}



/*

//============RESONANT VARIABLES==========
void GetResonantVar(int p, int q, OEPOS Q, double lprime, double omegabarprime, RVAR *r)
//Resonance (p+q)/p
//For external resonances q<0
//lprime,omegabar prime refer to secondary
{
if(q==0) return;
if(Q.modedisplay) OE_CalcValues(&Q);
double L=sqrt(rm1*Q.a);
double G=L*sqrt(1-Q.e*Q.e);
double H=G*cos(Q.i);
r->S=L-G;
r->Sz=G-H;
r->N=L*(p+q)/p - H;
double ll=(p+q)*lprime-p*Q.lamda;
double wwn,wwz,wws;
if(q>0)
{
wws=ll-q*Q.omegabar;
wwz=ll-q*Q.Omega;
wwn=-ll+q*omegabarprime;
}
else
{
wws=-ll+q*Q.omegabar;
wwz=-ll+q*Q.Omega;
wwn=ll-q*omegabarprime;
}
M180_180(&wws);M180_180(&wwz);M180_180(&wwn);
r->ws=wws;r->wz=wwz;r->wn=wwn;
}

//Resonant actions
void GetLGH(double a, double e, double ideg, double *L, double *G, double *H)
{
*L=sqrt(rm1*a);
*G=(*L)*sqrt(1-e*e);
*H=(*G)*cos(ideg*pi/180);
}


//=====================================================
//             ORBITAL ELEMENTS at t
//=====================================================

//1. Basic elliptic motion
OEPOS BasicEllipticMotion(OEPOS Q0, double t, int DISPLAYMODE)
//Q0:orbital elements at t=0;
//t: time for calculating orbital elements
//return orbital element at time t
{
if(Q0.modedisplay) OE_CalcValues(&Q0);
OEPOS Q=Q0; //elements a,e,i,omega,Omega,omegabar are constants
//Get M at time t
Q.M=Q0.resonance*t-Q0.M;
MPI_PI(&Q.M);
Q.lamda=Q.M+Q.omegabar; MPI_PI(&Q.lamda);
Q.E=GetEccentricAnomalyFromM(Q.M*pi/180,Q.e);MPI_PI(&Q.E);
Q.f=2*atan(sqrt((1+Q.e)/(1-Q.e))*tan(0.5*Q.E)); //MPI_PI(&Q.E); neeless
Q.tau=fmod(t,pi2/Q.resonance)-Q.M/Q.resonance;
if(DISPLAYMODE) OE_DisplayValues(&Q);
return Q;
}


*/


/*double GetMeanFromTrue(double f, double e)  //rad
//old version based on series expansion (not proper for large eccentricity values
{
//f: in rad
double ep=1;
double M=f;
ep*=e; M-=2*sin(f)*ep;
ep*=e; M+=3*sin(2*f)*ep/4;
ep*=e; M-=sin(3*f)*ep/3;
ep*=e; M+=(4*sin(2*f)+5*sin(4*f))*ep/32;
ep*=e; M+=(-5*sin(3*f)-3*sin(5*f))*ep/40;
ep*=e; M+=(9*sin(2*f)+18*sin(4*f)+7*sin(6*f))*ep/192;
ep*=e; M-=(7*sin(3*f)+7*sin(5*f)+2*sin(7*f))*ep/112;
ep*=e; M+=(24*sin(2*f)+60*sin(4*f)+40*sin(6*f)+9*sin(8*f))*ep/1024;
return M;
}*/